/*****************************************************************************
 *
 * grail - Gesture Recognition And Instantiation Library
 *
 * Copyright (C) 2010-2011 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include "v3/gesture.h"

#include <cstdio>
#include <stdexcept>

#include "utouch/grail.h"

#include "v3/event.h"
#include "v3/handle.h"
#include "v3/log.h"
#include "v3/recognizer.h"
#include "v3/slice.h"

namespace utouch {
namespace grail {

/**
 * @internal
 * Creates a new gesture
 */
Gesture::Gesture(Recognizer* recognizer, UGSubscription* subscription,
                 TouchSet& touches, uint64_t start_time)
    : recognizer_(recognizer),
      id_(recognizer_->handle()->NewGestureID(recognizer_)),
      subscription_(subscription),
      current_touches_(touches),
      unowned_touches_(touches),
      start_time_(start_time),
      owned_(false),
      not_owned_(false),
      recognized_(0),
      canceled_(false),
      ended_(false) {
}

/**
 * @internal
 * Copies an existing gesture, but with the passed in touch set
 */
Gesture::Gesture(const Gesture* other_gesture, TouchSet& touches)
    : recognizer_(other_gesture->recognizer_),
      id_(recognizer_->handle()->NewGestureID(recognizer_)),
      subscription_(other_gesture->subscription_),
      current_touches_(touches),
      unowned_touches_(touches),
      start_time_(other_gesture->start_time_),
      owned_(false),
      not_owned_(false),
      recognized_(other_gesture->recognized_),
      canceled_(false),
      ended_(false) {
}

/**
 * @internal
 * Add a new touch to an existing gesture
 */
void Gesture::AddTouch(UFTouchId touch_id) {
  current_touches_.insert(touch_id);
  unowned_touches_.insert(touch_id);
}

/**
 * @internal
 * Update a gesture with the passed in frame event and list of modified touches
 */
void Gesture::Update(UFEvent frame_event, TouchSet& modified_touches) {
  if (ended_) {
    /* The new event may have given us ownership of all the touches, in which
     * case we need to send all gesture slices to the client */
    if (IsActive())
      FlushSlices();

    /* If the gesture ended, no need to update state */
    return;
  }

  UFFrame frame;
  UFStatus status = frame_event_get_property(frame_event, UFEventPropertyFrame,
                                             &frame);
  if (status != UFStatusSuccess) {
    LOG(Warn) << "failed to get frame from event\n";
    return;
  }

  TouchSet ended_touches;
  bool touch_found = false;
  for (UFTouchId touch_id : modified_touches) {
    /* If the touch is not part of this gesture, skip it */
    if (current_touches_.find(touch_id) != current_touches_.end())
      touch_found = true;
    else
      continue;

    UFTouch touch;
    status = frame_frame_get_touch_by_id(frame, touch_id, &touch);
    if (status != UFStatusSuccess) {
      LOG(Warn) << "failed to get touch from frame\n";
      return;
    }

    UFTouchState state = frame_touch_get_state(touch);
    int owned;
    status = frame_touch_get_property(touch, UFTouchPropertyOwned, &owned);
    if (status != UFStatusSuccess) {
      LOG(Warn) << "failed to get touch from frame\n";
      return;
    }

    if (state == UFTouchStateEnd) {
      if (owned) {
        ended_touches.insert(touch_id);
      } else {
        /* If we receive a touch end before we own it, someone higher in the
         * stack accepted it. We need to cancel any handling of the touch. */
        canceled_ = true;
        break;
      }
    }
  }

  /* If the frame doesn't include any info about the touches in the gesture, or
   * the gesture has been canceled, bail. */
  if (!touch_found || canceled_)
    return;

  /* Is this the first slice of the gesture? */
  if (!last_slice_) {
    UGSlice* slice;
    /* Create gesture begin slice */
    slice = new UGSlice(shared_from_this(), frame_event, current_touches_,
                        recognized_);
    last_slice_ = SharedUGSlice(slice);
  } else {
    UGSlice* slice;
    /* Create gesture update or end slice */
    slice = new UGSlice(last_slice_, frame_event, current_touches_);
    last_slice_ = SharedUGSlice(slice);
  }

  slices_.push(last_slice_);

  for (UFTouchId touch_id : ended_touches)
    current_touches_.erase(touch_id);

  CheckOwned();

  /* Check for any new subscription matches */
  if (recognized_ != subscription_->mask())
    recognized_ |= last_slice_->CheckRecognition(*subscription_);

  /* Send slice events to client if gesture is active */
  if (IsActive())
    FlushSlices();

  /* If the gesture ended and nothing was recognized or all the gesture slices
   * have been sent, end the gesture */
  if (last_slice_->state() == UGGestureStateEnd &&
      (!recognized_ || IsConstructionFinished())) {
    ended_ = true;
    last_slice_.reset();
  }

  if (ended_ && !recognized_)
    canceled_ = true;
}

/**
 * @internal
 * Check if the gesture is active
 *
 * A gesture is active if a subscribed gesture has been matched and we own all
 * the touches in the gesture.
 */
bool Gesture::IsActive() const {
  return recognized_ && owned_;
}

/**
 * @internal
 * Check if construction is finished for the set of touches in the gesture
 *
 * See UGSlicePropertyConstructionFinished for details.
 */
bool Gesture::IsConstructionFinished() const {
  /* last_slice_ must either be valid or the gesture has ended, in which case
   * construction is considered finished. */
  return (!last_slice_ || last_slice_->construction_finished());
}

/**
 * @internal
 * Perform processing necessary for a gesture that has finished construction
 */
void Gesture::FinishConstruction() {
  if (!IsActive()) {
    LOG(Warn) << "attempted to finish construction of an inactive gesture\n";
    return;
  }

  /* If the gesture is already construction finished, bail */
  if (IsConstructionFinished())
    return;

  /* Create a new gesture slice to tell the client that construction has
   * finished */
  UGSlice* slice = new UGSlice(last_slice_);
  if (!slice)
    return;

  slice->set_construction_finished();

  last_slice_ = SharedUGSlice(slice);
  slices_.push(last_slice_);

  /* If construction has finished, we know we can send events to the client */
  FlushSlices();

  /* If we construction has finished on an ended gesture, we're done with it */
  if (last_slice_->state() == UGGestureStateEnd) {
    ended_ = true;
    last_slice_.reset();
  }
}

/**
 * @internal
 * Check that all the touches in the gesture are owned
 */
void Gesture::CheckOwned() {
  if (owned_ || not_owned_)
    return;

  for (auto it = unowned_touches_.begin(); it != unowned_touches_.end(); ) {
    UFTouchId touch_id = *it++;
    UFTouch touch;
    UFStatus status = frame_frame_get_touch_by_id(slices_.back()->frame(),
                                                  touch_id, &touch);
    if (status != UFStatusSuccess) {
      LOG(Warn) << "failed to get touch " << touch_id << " from frame, gesture "
                << id_ << " marked as not owned\n";
      not_owned_ = true;
      return;
    }

    int owned;
    status = frame_touch_get_property(touch, UFTouchPropertyOwned, &owned);
    if (status != UFStatusSuccess) {
      LOG(Warn) << "failed to get ownership property from touch, gesture "
                << id_ << " marked as not owned\n";
      not_owned_ = true;
      return;
    }

    if (owned) {
      LOG(Dbg) << "touch " << touch_id << " is now owned for gesture " << id_
               << "\n";
      unowned_touches_.erase(touch_id);
      continue;
    }

    if (frame_touch_get_state(touch) == UFTouchStateEnd) {
      LOG(Dbg) << "touch " << touch_id << "ended before it was owned, gesture "
               << id_ << " marked as not owned\n";
      not_owned_ = true;
    }

    return;
  }

  LOG(Dbg) << "all touches owned, marking gesture " << id_ << " as owned\n";
  owned_ = true;
  return;
}

/**
 * @internal
 * Flush all pending gesture slices to the client as grail events */
void Gesture::FlushSlices() {
  while (!slices_.empty()) {
    UGEvent* event = new UGEvent(slices_.front());
    slices_.front()->ReferenceGesture();
    recognizer_->handle()->EnqueueEvent(event);
    slices_.pop();
  }
}

/**
 * @internal
 * Check subscription timeouts
 */
uint64_t Gesture::Timeout() const {
  if (recognized_)
    return 0;

  uint64_t time = 0;
  if ((subscription_->mask() & UGGestureTypeDrag) &&
      subscription_->drag().timeout > time)
    time = subscription_->drag().timeout;
  if ((subscription_->mask() & UGGestureTypePinch) &&
      subscription_->pinch().timeout > time)
    time = subscription_->pinch().timeout;
  if ((subscription_->mask() & UGGestureTypeRotate) &&
      subscription_->rotate().timeout > time)
    time = subscription_->rotate().timeout;
  if ((subscription_->mask() & UGGestureTypeTap) &&
      subscription_->tap().timeout > time)
    time = subscription_->tap().timeout;

  return time;
}

/**
 * @internal
 * Get the current angle about the centroid of the gesture for a given touch */
float Gesture::AngleForTouch(UFTouchId touch_id) const {
  auto it = angles_.find(touch_id);
  if (it != angles_.end())
    return it->second;

  return 0;
}

/**
 * @internal
 * Set the angle about the centroid of the gesture for a given touch */
void Gesture::SetAngleForTouch(UFTouchId touch_id, float angle) {
  angles_[touch_id] = angle;
}

/**
 * @internal
 * Cancel the gesture
 */
void Gesture::Cancel() {
  while (!slices_.empty())
    slices_.pop();
  last_slice_.reset();
}

/**
 * @internal
 * End an active gesture */
void Gesture::End() {
  /* Create a new gesture slice in the end state */
  UGSlice* slice = new UGSlice(last_slice_, true);
  slices_.push(SharedUGSlice(slice));
  FlushSlices();
  last_slice_.reset();
  ended_ = true;
  current_touches_.clear();
  unowned_touches_.clear();
}

} // namespace grail
} // namespace utouch
