/*****************************************************************************
 *
 * grail - Gesture Recognition And Instantiation Library
 *
 * Copyright (C) 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

/**
 * @internal
 * @file Drag Threshold Test
 *
 * This test plays two different one-touch drags. The first drag is below the
 * threshold required for recognizing a drag gesture, and the second drag is
 * above the threshold.
 */

#include <map>
#include <memory>
#include <stdexcept>
#include <future>

#include <gtest/gtest.h>

#include "device.h"
#include "recording.h"
#include "x11/fixture.h"
#include "utouch/frame_x11.h"

using namespace utouch::grail::x11::testing;

class DragThreshold : public Test {
 public:
  DragThreshold() : device_(NULL), expect_drag_(false) {}

 protected:
  virtual void ProcessFrameEvents();
  virtual void ProcessGrailEvents();
  void Subscribe();
  void CheckSlice(UGSlice);

  UFDevice device_;
  bool expect_drag_;
};

void DragThreshold::ProcessFrameEvents() {
  UFEvent event;

  UFStatus status;
  while ((status = frame_get_event(frame_handle(), &event)) == UFStatusSuccess) {
    grail_process_frame_event(grail_handle(), event);

    if (frame_event_get_type(event) == UFEventTypeDeviceAdded) {
      UFDevice device;
      ASSERT_EQ(UFStatusSuccess,
                frame_event_get_property(event, UFEventPropertyDevice,
                                         &device));

      const char* name;
      ASSERT_EQ(UFStatusSuccess,
                frame_device_get_property(device, UFDevicePropertyName, &name));
      if (strcmp(name, "N-Trig-MultiTouch-Virtual-Device") == 0) {
        EXPECT_EQ(NULL, device_);
        device_ = device;
        Subscribe();
      }
    }

    frame_event_unref(event);
  }

  EXPECT_EQ(UFStatusErrorNoEvent, status);
}

void DragThreshold::ProcessGrailEvents() {
  UGEvent event;

  UGStatus status;
  while ((status = grail_get_event(grail_handle(), &event)) == UGStatusSuccess) {
    ASSERT_EQ(UGEventTypeSlice, grail_event_get_type(event));

    UGSlice slice;
    status = grail_event_get_property(event, UGEventPropertySlice, &slice);
    ASSERT_EQ(UGStatusSuccess, status);

    CheckSlice(slice);

    grail_event_unref(event);
  }

  EXPECT_EQ(UGStatusErrorNoEvent, status);
}

void DragThreshold::Subscribe() {
  UGSubscription subscription;
  UGStatus status = grail_subscription_new(&subscription);
  ASSERT_EQ(UGStatusSuccess, status);

  status = grail_subscription_set_property(subscription,
                                           UGSubscriptionPropertyDevice,
                                           &device_);
  ASSERT_EQ(UGStatusSuccess, status);

  const UFWindowId window_id =
      frame_x11_create_window_id(DefaultRootWindow(Display()));
  status = grail_subscription_set_property(subscription,
                                           UGSubscriptionPropertyWindow,
                                           &window_id);
  ASSERT_EQ(UGStatusSuccess, status);

  const UGGestureTypeMask mask = UGGestureTypeDrag;
  status = grail_subscription_set_property(subscription,
                                           UGSubscriptionPropertyMask,
                                           &mask);
  ASSERT_EQ(UGStatusSuccess, status);

  const unsigned int touches = 1;
  status = grail_subscription_set_property(subscription,
                                           UGSubscriptionPropertyTouchesStart,
                                           &touches);
  ASSERT_EQ(UGStatusSuccess, status);

  status = grail_subscription_set_property(subscription,
                                           UGSubscriptionPropertyTouchesMinimum,
                                           &touches);
  ASSERT_EQ(UGStatusSuccess, status);

  status = grail_subscription_activate(grail_handle(), subscription);
  ASSERT_EQ(UGStatusSuccess, status);
}

void DragThreshold::CheckSlice(UGSlice slice) {
  if (grail_slice_get_state(slice) != UGGestureStateEnd)
    return;

  if (expect_drag_)
    EXPECT_TRUE(grail_slice_get_recognized(slice) & UGGestureTypeDrag);
  else
    EXPECT_FALSE(grail_slice_get_recognized(slice) & UGGestureTypeDrag);
}

TEST_F(DragThreshold, Recording) {
  utouch::evemu::Device device("recordings/ntrig-dell-xt2.prop");

  /* Pump once to ensure the X server has initialized the device */
  PumpEvents();
  ASSERT_NE(nullptr, device_) << "X server failed to initialize touchscreen";

  expect_drag_ = false;

  utouch::evemu::Recording bad_drag(device,
                                    "recordings/ntrig-drag-below-thresh.event");

  /* We use the c++11 future module so any exceptions thrown by the thread can
   * be caught later on. If we used the thread module, exceptions would take the
   * whole thing down. */
  std::future<void> future = std::async(std::launch::async,
                                        &utouch::evemu::Recording::Play,
                                        &bad_drag);

  PumpEvents();

  future.wait();

  expect_drag_ = true;

  utouch::evemu::Recording good_drag(
      device,
      "recordings/ntrig-drag-above-thresh.event");

  future = std::async(std::launch::async, &utouch::evemu::Recording::Play,
                      &good_drag);

  PumpEvents();

  future.wait();
}
